//-------------------------------------------------------------------------
// 
// Copyright Institut Telecom 2008, 2009
// Licensed Materials, Program Property of Institut Telecom
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  Institut Telecom 
// retains full right to use the code for his/her own purpose, assign or donate 
// the code to a third party and to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Khaled Mammou
//
//
//    http://www-artemis.it-sudparis.eu/
//    
//    Khaled Mammou      (khaled_mamou@yahoo.fr)
//
//-------------------------------------------------------------------------
#include "TFANGeometryEncoder.h"
#include "arithmetic_codec.h"
#include "BinaryIO.h"
#include "MathTools.h"
#include "utils.h"
#include <math.h>

void TFANGeometryEncoder::Init()
{
	qTable_ = new int [m_nVertices * m_dim];
	errTable_= new int [m_nVertices * m_dim];
	tagVertex_ = new int [m_nVertices];
	for(int v = 0; v < m_nVertices; v++) {
		tagVertex_[v] = 0;
		for (int d = 0; d < m_dim; d++) {
			qTable_[v*m_dim+d] = 0;
			errTable_[v*m_dim+d] = 0;
		}
	}
	tagTriangle_ = new int [m_nTriangles];
	for(int t = 0; t < m_nTriangles; t++) {
		tagTriangle_[t] = 0;
	}
	min_ = new float[m_dim];
	max_ = new float[m_dim];
	for (int d = 0; d < m_dim; d++) {
		min_[d] = 0.0f;
		max_[d] = 0.0f;
	}
}

bool TFANGeometryEncoder::GetCoordIndex(int pos, int * coordIndex) {
	if (pos < m_nTriangles) {
		for(int h = 0; h < 3; h++) {
			coordIndex[h] = m_triangles[pos*3+h];
		}
		return true;
	}
	return false;
}
void TFANGeometryEncoder::AddNeighborVertex2Vertex(int v1, int v2)
{
	int found = 0;	
	for (IntVect::iterator posc = vertex2Vertex_[v1].begin(); posc != vertex2Vertex_[v1].end(); ++posc) {
		if ((*posc) == v2) {
			found = 1;
			break;
		}
	}
	if (found == 0)
		vertex2Vertex_[v1].push_back(v2);
}

void TFANGeometryEncoder::ComputeVertex2Vertex(){
	for (int f = 0; f < m_nTriangles; f++) {
		AddNeighborVertex2Vertex(m_triangles[f*3+0], m_triangles[f*3+1]);
		AddNeighborVertex2Vertex(m_triangles[f*3+0], m_triangles[f*3+2]);

		AddNeighborVertex2Vertex(m_triangles[f*3+1], m_triangles[f*3+0]);
		AddNeighborVertex2Vertex(m_triangles[f*3+1], m_triangles[f*3+2]);

		AddNeighborVertex2Vertex(m_triangles[f*3+2], m_triangles[f*3+0]);
		AddNeighborVertex2Vertex(m_triangles[f*3+2], m_triangles[f*3+1]);
	}
}
void TFANGeometryEncoder::AddNeighborVertex2Triangle(int v, int t)
{
	int found = 0;	
	for (IntVect::iterator posc = vertex2Triangle_[v].begin(); posc != vertex2Triangle_[v].end(); ++posc) {
		if ((*posc) == t) {
			found = 1;
			break;
		}
	}
	if (found == 0)
		vertex2Triangle_[v].push_back(t);
}

void TFANGeometryEncoder::ComputeVertex2Triangle(){
	for (int f = 0; f < m_nTriangles; f++) {
		AddNeighborVertex2Triangle(m_triangles[f*3+0], f);
		AddNeighborVertex2Triangle(m_triangles[f*3+1], f);
		AddNeighborVertex2Triangle(m_triangles[f*3+2], f);
	}
}
void TFANGeometryEncoder::AddNeighborTriangle2Triangle(int t1, int t2)
{
	int found = 0;	
	for (IntVect::iterator posc = triangle2Triangle_[t1].begin(); posc != triangle2Triangle_[t1].end(); ++posc) {
		if ((*posc) == t2) {
			found = 1;
			break;
		}
	}
	if (found == 0)
		triangle2Triangle_[t1].push_back(t2);
}

void TFANGeometryEncoder::ComputeTriangle2Triangle(){
	int coordIndex[3] = {-1, -1, -1};

	for (int t = 0; t < m_nTriangles; t++) {
		GetCoordIndex(t, coordIndex);
		for (int k = 0; k < 3; k++) {
			for (IntVect::iterator pt0 = vertex2Triangle_[coordIndex[k]].begin(); pt0 != vertex2Triangle_[coordIndex[k]].end(); ++pt0){
				if ((*pt0) != t){
					int tr0 = (*pt0);
					for (int i=0; i <  (int) vertex2Triangle_[coordIndex[(k+1)%3]].size();i++){
						int tr1 = vertex2Triangle_[coordIndex[(k+1)%3]][i];
						if (tr1 == tr0){
							AddNeighborTriangle2Triangle(t, tr1);
						}
					}
				}
			}
		}
	}
}

int TFANGeometryEncoder::Round(double a){
      if (a >= 0.0) return (int) (a+0.5);
      return -(int) (-a+0.5);
}

void TFANGeometryEncoder::Quantize(){
	for (int d = 0; d < m_dim; d++){  
		min_[d] = m_geometry[d];
		max_[d] = m_geometry[d];
	}
    for (int v = 0; v < m_nVertices; v++) {
            for (int d = 0; d < m_dim; d++){
                  if ( m_geometry[v*m_dim+d] >  max_[d])  max_[d] = m_geometry[v*m_dim+d];
                  if ( m_geometry[v*m_dim+d] <  min_[d])  min_[d] = m_geometry[v*m_dim+d];
            }
    }
    float mm = max_[0] - min_[0];
	for (int d = 0; d < m_dim; d++){  
		if ( (max_[d] - min_[d]) > mm) mm =  max_[d] - min_[d];
	}
	for (int d = 0; d < m_dim; d++){
		max_[d] =  min_[d] + mm;
	}    

	
	// Constant for uniform quantization
    double * delta = new double[m_dim];
	for (int d = 0; d < m_dim; d++){  
		delta[d] = 1.0;
		if ((max_[d]-min_[d]) != 0.0) delta[d] = (double) ((1 << m_nQBits) - 1) / (max_[d]-min_[d]);
	}

	for (int v = 0; v < m_nVertices; v++) {
	  for (int d = 0; d < m_dim; d++){
		qTable_[v*m_dim+d] = (int) Round((m_geometry[v*m_dim+d] - min_[d]) * delta[d]);
      }
	}
	delete [] delta;
}
void TFANGeometryEncoder::GetVisitedNeighbors(int v, IntVect & vN){
	for (int i = 0; i!= vertex2Vertex_[v].size(); i++) {
		if (tagVertex_[vertex2Vertex_[v][i]] != 0) {
			vN.push_back(vertex2Vertex_[v][i]);
		}
	}
}
int TFANGeometryEncoder::BaryPred(){
	ComputeVertex2Vertex();
	for(int v = 0; v < m_nVertices; v++) {
		BaryPredVertex(v);
	}
	return 0;
}
int TFANGeometryEncoder::LPararPred(int a, int b, int c, int d){
	if (tagVertex_[d] == 0) {
		tagVertex_[d] = 1;
		if (m_dim != 3) {
			for (int i = 0; i < m_dim; i++) {
				errTable_[d*m_dim+i] = qTable_[d*m_dim+i] + qTable_[a*m_dim+i] - qTable_[b*m_dim+i] - qTable_[c*m_dim+i];
			}
		}
		else {
			float pa[3] = {(float) qTable_[a*m_dim+0], (float) qTable_[a*m_dim+1], (float) qTable_[a*m_dim+2]};
			float pb[3] = {(float) qTable_[b*m_dim+0], (float) qTable_[b*m_dim+1], (float) qTable_[b*m_dim+2]};
			float pc[3] = {(float) qTable_[c*m_dim+0], (float) qTable_[c*m_dim+1], (float) qTable_[c*m_dim+2]};
			float pd[3] = {(float) qTable_[d*m_dim+0], (float) qTable_[d*m_dim+1], (float) qTable_[d*m_dim+2]};

			float u[3] = {0.0f, 0.0f, 0.0f};
			MathTools::vecteur(pb, pc, u, 3);
			MathTools::vectorUnitary(u);

			float v[3] = {0.0f, 0.0f, 0.0f};
			for (int k = 0; k <3; k++) {
				v[k] = 0.5f * (pb[k] + pc[k]) - pa[k];
			}
			float pdot = MathTools::vectorDot(v, u);
			for (int k = 0; k <3; k++) {
				v[k] -= pdot * u[k];
			}
			MathTools::vectorUnitary(v);

			float w[3] = {0.0f, 0.0f, 0.0f};
			MathTools::vectoriel(u, v, w);

			float err[3] = {0.0f, 0.0f, 0.0f};
			MathTools::vecteur(pb, pd, err, 3);

			float errL[3] = {0.0f, 0.0f, 0.0f};
			errL[0] = (float) Round(MathTools::vectorDot(err, u));
			errL[1] = (float) Round(MathTools::vectorDot(err, v));
			errL[2] = (float) Round(MathTools::vectorDot(err, w));

			float pdR[3] = {0.0f, 0.0f, 0.0f};
			pdR[0] = (float) Round(pb[0] + errL[0] * u[0] + errL[1] * v[0] +errL[2] * w[0]);
			pdR[1] = (float) Round(pb[1] + errL[0] * u[1] + errL[1] * v[1] +errL[2] * w[1]);
			pdR[2] = (float) Round(pb[2] + errL[0] * u[2] + errL[1] * v[2] +errL[2] * w[2]);
			
			for (int i = 0; i <3; i++) {
				qTable_[d*m_dim+i] = (int) pdR[i];
				errTable_[d*m_dim+i] = (int) errL[i];
			}


		}
	}
	return 0;
}
void TFANGeometryEncoder::BaryPredVertex(int v){
	if (tagVertex_[v] == 0) {
		IntVect vN;
		GetVisitedNeighbors(v, vN);
		if (vN.size() == 0) {
			tagVertex_[v] = 2;
			for (int i = 0; i < m_dim; i++) {
				errTable_[v*m_dim+i] = qTable_[v*m_dim+i];
			}
		}
		else {
			tagVertex_[v] = 1;
			for (int i = 0; i < m_dim; i++) {
				double bary = 0.0;
				for (int k = 0; k != vN.size(); k++) {
					bary += qTable_[vN[k]*m_dim+i];
				}
				bary/= vN.size();
				errTable_[v*m_dim+i] = qTable_[v*m_dim+i] - Round(bary);
			}
		}
	}
}

int TFANGeometryEncoder::DeltaPred(){
	for(int v = 0; v < m_nVertices; v++) {
		if ( tagVertex_[v] == 0) {
			if ( v==0) {
				tagVertex_[v] = 2;
				for (int i = 0; i < m_dim; i++) {
					errTable_[v*m_dim+i] = qTable_[v*m_dim+i];
				}
			}
			else {
				tagVertex_[v] = 1;
				for (int i = 0; i < m_dim; i++) {
					errTable_[v*m_dim+i] = qTable_[v*m_dim+i] - qTable_[(v-1)*m_dim+i];
				}
			}
		}
	}
	return 0;
}
int TFANGeometryEncoder::ParaPred(){
	ComputeVertex2Vertex();
	ComputeVertex2Triangle();
	ComputeTriangle2Triangle();
	int currentT[3] = {-1, -1, -1};
	int nextT[3] = {-1, -1, -1};
	for(int t = 0; t != m_nTriangles; t++) {
		if (tagTriangle_[t] == 0) {
			GetCoordIndex(t, currentT);
			for (int k = 0; k != 3; k++) {
				BaryPredVertex(currentT[k]);
			}
			IntVect LT;
			LT.push_back(t);
			tagTriangle_[t] = 1;
			int ct = -1;
			int nt = -1;
			while (LT.size() != 0){
				ct = LT[0]; LT.erase(LT.begin());
				GetCoordIndex(ct, currentT);
				for (int u = 0; u != triangle2Triangle_[ct].size(); u++){
					nt = triangle2Triangle_[ct][u];
					if (tagTriangle_[nt] == 0) {
						LT.push_back(nt);
						tagTriangle_[nt] = 1;
						GetCoordIndex(nt, nextT);
						// determin the commun edge
						IntVect cEdge;
						for (int k1 = 0; k1 != 3; k1++){
							for (int k2 = 0; k2 != 3; k2++){
								if (nextT[k2] == currentT[k1]) cEdge.push_back(nextT[k2]);
							}
						}


						// determin a and d
						int a = -1;
						int b = cEdge[0];
						int c = cEdge[1];
						int d = -1;
						for (int k1 = 0; k1 != 3; k1++){
							if ((currentT[k1] != b) && (currentT[k1] != c)) a = currentT[k1];
							if ((nextT[k1] != b) && (nextT[k1] != c)) d = nextT[k1];
						}

						if ( tagVertex_[d] == 0) {
							
//							LPararPred(a, b, c, d);
							for (int i = 0; i < m_dim; i++) {
								errTable_[d*m_dim+i] = qTable_[d*m_dim+i] + qTable_[a*m_dim+i] - qTable_[b*m_dim+i] - qTable_[c*m_dim+i];
							}
/*
							printf("[%i]\t", d);
							for (int i = 0; i < m_dim; i++) {
								printf("%i\t", errTable_[d*m_dim+i]);
							}
							printf("\n");
*/
							tagVertex_[d] = 3;
						}
					}
				}
			}
		}
	}
	// for isolated vertices
	for(int v = 0; v < m_nVertices; v++) {
		BaryPredVertex(v);
	}
	return 0;
}
int TFANGeometryEncoder::AdapPred(){
	BinaryIO stream;
	ComputeVertex2Vertex();
	ComputeVertex2Triangle();
	ComputeTriangle2Triangle();
	int currentT[3] = {-1, -1, -1};
	int nextT[3] = {-1, -1, -1};
	for(int t = 0; t != m_nTriangles; t++) {
		if (tagTriangle_[t] == 0) {
			GetCoordIndex(t, currentT);
			for (int k = 0; k != 3; k++) {
				BaryPredVertex(currentT[k]);
			}
			IntVect LT;
			LT.push_back(t);
			tagTriangle_[t] = 1;
			int ct = -1;
			int nt = -1;
			while (LT.size() != 0){
				ct = LT[0]; LT.erase(LT.begin());
				GetCoordIndex(ct, currentT);
				for (int u = 0; u != triangle2Triangle_[ct].size(); u++){
					nt = triangle2Triangle_[ct][u];
					if (tagTriangle_[nt] == 0) {
						LT.push_back(nt);
						tagTriangle_[nt] = 1;
						GetCoordIndex(nt, nextT);
						// determin the commun edge
						IntVect cEdge;
						for (int k1 = 0; k1 != 3; k1++){
							for (int k2 = 0; k2 != 3; k2++){
								if (nextT[k2] == currentT[k1]) cEdge.push_back(nextT[k2]);
							}
						}


						// determin a and d
						int a = -1;
						int b = cEdge[0];
						int c = cEdge[1];
						int d = -1;
						for (int k1 = 0; k1 != 3; k1++){
							if ((currentT[k1] != b) && (currentT[k1] != c)) a = currentT[k1];
							if ((nextT[k1] != b) && (nextT[k1] != c)) d = nextT[k1];
						}

						if ( tagVertex_[d] == 0) {
							stream.Clear();
							tagVertex_[d] = 0;
							BaryPredVertex(d);
							for (int i = 0; i < m_dim; i++) {
								stream.ExpGolombEncode(abs(errTable_[d*m_dim+i]), 0);
								stream.WriteBin(0); // signs
							}
							int errB = stream.Size();

							
							stream.Clear();
							tagVertex_[d] = 0;
							for (int i = 0; i < m_dim; i++) {
								errTable_[d*m_dim+i] = qTable_[d*m_dim+i] + qTable_[a*m_dim+i] - qTable_[b*m_dim+i] - qTable_[c*m_dim+i];
								stream.ExpGolombEncode(abs(errTable_[d*m_dim+i]), 0);
								stream.WriteBin(0); // signs
							}
							int errP = stream.Size();


							stream.Clear();

							int errA = m_nQBits*m_dim;

							tagVertex_[d] = 0;
							if (errB < errP) {// barycenter
								BaryPredVertex(d);
								vertexPredMode_.push_back(1);
							}
							else {
								for (int i = 0; i < m_dim; i++) {
									errTable_[d*m_dim+i] = qTable_[d*m_dim+i] + qTable_[a*m_dim+i] - qTable_[b*m_dim+i] - qTable_[c*m_dim+i];
								}
								tagVertex_[d] = 1;						
								vertexPredMode_.push_back(0);
							}
							tagVertex_[d] = 1;
						}
					}
				}
			}
		}
	}
	// for isolated vertices
	for(int v = 0; v < m_nVertices; v++) {
		BaryPredVertex(v);
	}
	return 0;
}
int TFANGeometryEncoder::FastParaPred(){
	int a = 0;
	int b = 0;
	int c = 0;
	int d = 0;
	int e = -1;
	int ta = -1;
	int x = -1;
	int y = -1;
	int z = -1;
	for(int v = 0; v < tfans_->GetV(); v++) {
		if(tfans_->GetSize(v)>2){
			// first vertex
			a = tfans_->GetElement(v,0);
			if(tagVertex_[a]==0){
				for (int i = 0; i < m_dim; i++) {
					errTable_[a*m_dim+i] = qTable_[a*m_dim+i];
				}
				tagVertex_[a]=2;
			}

			// second vertex
			b = tfans_->GetElement(v,1);
			if(tagVertex_[b]==0){
				for (int i = 0; i < m_dim; i++) {
					errTable_[b*m_dim+i] = qTable_[b*m_dim+i] - qTable_[a*m_dim+i];
				}
				tagVertex_[b]=1;
			}

			//third vertex
			c = tfans_->GetElement(v,2);
			if(tagVertex_[c]==0){
				e = -1;
				ta = -1;
				x = -1;
				y = -1;
				z = -1;

				for (int u = 0; u < V2T_->GetSize(a); u++) {
					ta = V2T_->GetElement(a, u);
					x = m_triangles[ta*3];
					y = m_triangles[ta*3+1];
					z = m_triangles[ta*3+2];
					if ( (x==a) && (y==b) && (z!=c) && (tagVertex_[z]!=0) ) { e = z; break;}
					if ( (x==a) && (z==b) && (y!=c) && (tagVertex_[y]!=0) ) { e = y; break;}

					if ( (y==a) && (x==b) && (z!=c) && (tagVertex_[z]!=0) ) { e = z; break;}
					if ( (y==a) && (z==b) && (x!=c) && (tagVertex_[x]!=0) ) { e = x; break;}

					if ( (z==a) && (x==b) && (y!=c) && (tagVertex_[y]!=0) ) { e = y; break;}
					if ( (z==a) && (y==b) && (x!=c) && (tagVertex_[x]!=0) ) { e = x; break;}
				}
				if (e==-1) {
					for (int i = 0; i < m_dim; i++) {
						errTable_[c*m_dim+i] = qTable_[c*m_dim+i] - qTable_[b*m_dim+i];
					}
					tagVertex_[c]=1;
				}
				else {
					for (int i = 0; i < m_dim; i++) {
						errTable_[c*m_dim+i] = qTable_[c*m_dim+i] - qTable_[b*m_dim+i] + qTable_[e*m_dim+i] - qTable_[a*m_dim+i];
					}
					tagVertex_[c]=3;
				}
			}

			// remaining vertices
			for(int k = 3; k < tfans_->GetSize(v); k++){
				d = tfans_->GetElement(v,k);
				if(tagVertex_[d]==0){
					for (int i = 0; i < m_dim; i++) {
						errTable_[d*m_dim+i] = qTable_[d*m_dim+i] - qTable_[c*m_dim+i] + qTable_[b*m_dim+i] - qTable_[a*m_dim+i];
					}
					b = c;
					c = d;
					tagVertex_[d]=3;
				}
			}
		}
	}
	for (int v=0; v < m_nVertices; v++) {
		if(tagVertex_[v]==0){
			for (int i = 0; i < m_dim; i++) {
				errTable_[v*m_dim+i] = qTable_[v*m_dim+i];
			}
			tagVertex_[v]=2;
		}
	}
	return 0;
}


int TFANGeometryEncoder::Compute(){
	Quantize();
	switch(m_predMode) {
		case -1 :		// nothing
			for(int v = 0; v < m_nVertices; v++) {
				tagVertex_[v] = 2;
				for (int i = 0; i < m_dim; i++) {
					errTable_[v*m_dim+i] = qTable_[v*m_dim+i];
				}
			}

		break;
		case 0 :		// barycenter
			BaryPred();
		break;
		
		case 1 :		// delta
			DeltaPred();
		break;

		case 2 :		// para
			ParaPred();
		break;

		case 3 :		// Adaptive
			AdapPred();
		break;
		case 4 :		// Fast para
			FastParaPred();
		break;
	}
/*
	printf("Encoder\n");

	for(int v = 0; v < vertexPredMode_.size(); v++) {
		printf("%i\n", vertexPredMode_[v]);
	}


	printf("Trivial path\n");


	for(int v = 0; v < m_nVertices; v++) {
		printf("%i\t(%i)\t", v, tagVertex_[v]);
		for (int i = 0; i < m_dim; i++) {
			printf("%i\t", errTable_[v*m_dim+i]);
		}
		printf("\n");
	}

/*
	EntropyCalc entropy(errTable_, m_nVertices*m_dim);
	int size = 0;
	entropy.ComputeEntropy(size);
	printf("gemetry (bpv) %2.3f\n", ((float) size)/m_nVertices); 
*/
	return 0;
}

int TFANGeometryEncoder::Encode(unsigned char * &compressedStream, unsigned int & compressedStreamSize)
{
	Init();
	Compute();
	int size = 0;
	unsigned char K = 0;
	unsigned char M = 0;
	GetEncodeACMParameters(M, K);
	unsigned char encodingMode = 0;
	if (EncodeRLE(M,K) < EncodeNoRLE(M,K)) {
		encodingMode = 1;	// RLE
	}
	else {
		encodingMode = 0; // No RLE
	}
	Write2Stream(compressedStream, compressedStreamSize, &encodingMode, sizeof(unsigned char));

	unsigned char q = (unsigned char) m_nQBits;
	unsigned char predMode = (unsigned char) m_predMode;

//	Write2Stream(compressedStream, compressedStreamSize, &q, sizeof(unsigned char));
	Write2Stream(compressedStream, compressedStreamSize, &predMode, sizeof(unsigned char));
	Write2Stream(compressedStream, compressedStreamSize, &M, sizeof(unsigned char));
	Write2Stream(compressedStream, compressedStreamSize, &K, sizeof(unsigned char));
	Write2Stream(compressedStream, compressedStreamSize, min_, sizeof(float) * m_dim);
	Write2Stream(compressedStream, compressedStreamSize, max_, sizeof(float) * m_dim);

	size += 5 + 2 * m_dim*4;

	unsigned int NMAX = (int) m_nVertices * m_dim * 4;
	Adaptive_Bit_Model bModelPredMode;
	Adaptive_Data_Model mModel(2*M+3);
	Arithmetic_Codec ace(NMAX, NULL);
	Adaptive_Bit_Model * repModel = new Adaptive_Bit_Model[m_dim];
	ace.start_encoder();
	if (m_predMode == 3) {
		unsigned int nbrPredModes = (unsigned int) vertexPredMode_.size();
		Write2Stream(compressedStream, compressedStreamSize, &nbrPredModes, sizeof(unsigned int) * m_dim);
		size += 4;
		for(int v = 0; v < (int) nbrPredModes; v++) {
			ace.encode(vertexPredMode_[v], bModelPredMode);
		}
	}
/*
	FILE * fp = fopen("Geometry.txt", "w");
	for(int v = 0; v < m_nVertices; v++) {
		for (int i = 0; i < m_dim; i++) {
			fprintf(fp, "%i\n", errTable_[v*m_dim+i]);
		}
	}
	fclose(fp);
*/

	if ( encodingMode == 1) { // With RLE
		for(int v = 0; v < m_nVertices; v++) {
			for (int i = 0; i < m_dim; i++) {
				if ((v != 0) &&( errTable_[v*m_dim+i] == errTable_[(v-1)*m_dim+i])){
					ace.encode(0, repModel[i]);
				}
				else {
					ace.encode(1, repModel[i]);
					if (abs(errTable_[v*m_dim+i]) <= M) {
						ace.encode(errTable_[v*m_dim+i] + M, mModel);
					}
					else {
						if (errTable_[v*m_dim+i] < 0) {
							ace.encode(2*M+1, mModel);
						} 
						else {
							ace.encode(2*M+2, mModel);
						}
						ace.ExpGolombEncode(abs(errTable_[v*m_dim+i])-M-1, K);
					}
				}
			}
		}
	}
	else {	// No RLE
		for(int v = 0; v < m_nVertices; v++) {
			for (int i = 0; i < m_dim; i++) {
				if (abs(errTable_[v*m_dim+i]) <= M) {
					ace.encode(errTable_[v*m_dim+i] + M, mModel);
				}
				else {
					if (errTable_[v*m_dim+i] < 0) {
						ace.encode(2*M+1, mModel);
					} 
					else {
						ace.encode(2*M+2, mModel);
					}
					ace.ExpGolombEncode(abs(errTable_[v*m_dim+i])-M-1, K);
				}
			}
		}
	}
	size += (int) ace.write_to_mem(compressedStream, compressedStreamSize);
	delete [] repModel;
	return size;
}
int TFANGeometryEncoder::EncodeRLE(unsigned char M, unsigned char K){
	int size = 0;
	unsigned int NMAX = (int) m_nVertices * m_dim * 4;
	Adaptive_Data_Model mModelRLE(2*M+3);
	Arithmetic_Codec aceRLE(NMAX, NULL);
	Adaptive_Bit_Model * repModelRLE = new Adaptive_Bit_Model[m_dim];
	aceRLE.start_encoder();

	for(int v = 0; v < m_nVertices; v++) {
		for (int i = 0; i < m_dim; i++) {
			if ((v != 0) &&( errTable_[v*m_dim+i] == errTable_[(v-1)*m_dim+i])){
				aceRLE.encode(0, repModelRLE[i]);
			}
			else {
				aceRLE.encode(1, repModelRLE[i]);
				if (abs(errTable_[v*m_dim+i]) <= M) {
					aceRLE.encode(errTable_[v*m_dim+i] + M, mModelRLE);
				}
				else {
					if (errTable_[v*m_dim+i] < 0) {
						aceRLE.encode(2*M+1, mModelRLE);
					} 
					else {
						aceRLE.encode(2*M+2, mModelRLE);
					}
					aceRLE.ExpGolombEncode(abs(errTable_[v*m_dim+i])-M-1, K);
				}
			}
		}
	}
	delete [] repModelRLE;
	size = (int) aceRLE.stop_encoder();
//	printf("RLE Geometry ACM (%i, %i) -> %2.3f\n", M, K,  (float) size*8/m_nVertices);
	return size;
}
int TFANGeometryEncoder::EncodeNoRLE(unsigned char M, unsigned char K){
	int size = 0;
	unsigned int NMAX = (int) m_nVertices * m_dim * 4;
	Adaptive_Data_Model mModel(2*M+3);
	Arithmetic_Codec ace(NMAX, NULL);
	ace.start_encoder();

	for(int v = 0; v < m_nVertices; v++) {
		for (int i = 0; i < m_dim; i++) {
			if (abs(errTable_[v*m_dim+i]) <= M) {
				ace.encode(errTable_[v*m_dim+i] + M, mModel);
			}
			else {
				if (errTable_[v*m_dim+i] < 0) {
					ace.encode(2*M+1, mModel);
				} 
				else {
					ace.encode(2*M+2, mModel);
				}
				ace.ExpGolombEncode(abs(errTable_[v*m_dim+i])-M-1, K);
			}
		}
	}
	size = (int) ace.stop_encoder();
//	printf("No RLE Geometry ACM (%i, %i) -> %2.3f\n", M, K,  (float) size*8/m_nVertices);
	return size;
}
int TFANGeometryEncoder::EncodeACM(unsigned char M, unsigned char K){
	unsigned int NMAX = (int) m_nVertices * m_dim * 16;
	Adaptive_Bit_Model bModelPredMode;
	Adaptive_Bit_Model bModelSign;
	Adaptive_Data_Model mModel(2*M+3);
	Arithmetic_Codec ace(NMAX, NULL);
	ace.start_encoder();
	int size = 0;

	for(int v = 0; v < m_nVertices; v++) {
		for (int i = 0; i < m_dim; i++) {
			if (abs(errTable_[v*m_dim+i]) <= M) {
				ace.encode(errTable_[v*m_dim+i] + M, mModel);
			}
			else {
				if (errTable_[v*m_dim+i] < 0) {
					ace.encode(2*M+1, mModel);
				} 
				else {
					ace.encode(2*M+2, mModel);
				}
				ace.ExpGolombEncode(abs(errTable_[v*m_dim+i])-M-1, K);
			}			
		}
	}
	size += (int) ace.stop_encoder();
//	printf("	Geometry ACM (%i, %i) -> %2.3f\n", M, K,  (float) size*8/m_nVertices);
	return size;
}
void TFANGeometryEncoder::GetEncodeACMParameters(unsigned char & M, unsigned char & K){
	int size = 0;
	int bestSize = 0;
	K = 0;
	M = 0;
	for (int k = 0; k != 16; k++) {
		for (int m = 0; m != 255; m++) {
			size = EncodeACM(m, k);
			if (((m==0) && (k==0)) || (bestSize > size)) {
				bestSize = size;
				K = (unsigned char) k;
				M = (unsigned char) m;
			}
		}
	}
}